#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License
#
# Copyright 2012 Sony Mobile Communications. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Unit tests for the Pygerrit2 helper methods."""

import re
import unittest

from mock import patch
from pygerrit2 import GerritReviewMessageFormatter, GerritReview
from pygerrit2 import HTTPBasicAuthFromNetrc, HTTPDigestAuthFromNetrc, Anonymous
from pygerrit2 import GerritRestAPI

EXPECTED_TEST_CASE_FIELDS = ["header", "footer", "paragraphs", "result"]


TEST_CASES = [
    {"header": None, "footer": None, "paragraphs": [], "result": ""},
    {"header": "Header", "footer": "Footer", "paragraphs": [], "result": ""},
    {"header": None, "footer": None, "paragraphs": ["Test"], "result": "Test"},
    {
        "header": None,
        "footer": None,
        "paragraphs": ["Test", "Test"],
        "result": "Test\n\nTest",
    },
    {
        "header": "Header",
        "footer": None,
        "paragraphs": ["Test"],
        "result": "Header\n\nTest",
    },
    {
        "header": "Header",
        "footer": None,
        "paragraphs": ["Test", "Test"],
        "result": "Header\n\nTest\n\nTest",
    },
    {
        "header": "Header",
        "footer": "Footer",
        "paragraphs": ["Test", "Test"],
        "result": "Header\n\nTest\n\nTest\n\nFooter",
    },
    {
        "header": "Header",
        "footer": "Footer",
        "paragraphs": [["One"]],
        "result": "Header\n\n* One\n\nFooter",
    },
    {
        "header": "Header",
        "footer": "Footer",
        "paragraphs": [["One", "Two"]],
        "result": "Header\n\n* One\n* Two\n\nFooter",
    },
    {
        "header": "Header",
        "footer": "Footer",
        "paragraphs": ["Test", ["One"], "Test"],
        "result": "Header\n\nTest\n\n* One\n\nTest\n\nFooter",
    },
    {
        "header": "Header",
        "footer": "Footer",
        "paragraphs": ["Test", ["One", "Two"], "Test"],
        "result": "Header\n\nTest\n\n* One\n* Two\n\nTest\n\nFooter",
    },
    {
        "header": "Header",
        "footer": "Footer",
        "paragraphs": ["Test", "Test", ["One"]],
        "result": "Header\n\nTest\n\nTest\n\n* One\n\nFooter",
    },
    {
        "header": None,
        "footer": None,
        "paragraphs": [["* One", "* Two"]],
        "result": "* One\n* Two",
    },
    {
        "header": None,
        "footer": None,
        "paragraphs": [["*    One  ", "    * Two  "]],
        "result": "* One\n* Two",
    },
    {"header": None, "footer": None, "paragraphs": [["*", "*"]], "result": ""},
    {"header": None, "footer": None, "paragraphs": [["", ""]], "result": ""},
    {"header": None, "footer": None, "paragraphs": [["  ", "  "]], "result": ""},
    {
        "header": None,
        "footer": None,
        "paragraphs": [["* One", "  ", "* Two"]],
        "result": "* One\n* Two",
    },
]


class TestGerritReviewMessageFormatter(unittest.TestCase):
    """Test that the GerritReviewMessageFormatter class behaves properly."""

    def _check_test_case_fields(self, test_case, i):
        for field in EXPECTED_TEST_CASE_FIELDS:
            self.assertTrue(
                field in test_case,
                "field '%s' not present in test case #%d" % (field, i),
            )
        self.assertTrue(
            isinstance(test_case["paragraphs"], list),
            "'paragraphs' field is not a list in test case #%d" % i,
        )

    def test_is_empty(self):
        """Test if message is empty for missing header and footer."""
        fmt = GerritReviewMessageFormatter(header=None, footer=None)
        self.assertTrue(fmt.is_empty())
        fmt.append(["test"])
        self.assertFalse(fmt.is_empty())

    def test_message_formatting(self):
        """Test message formatter for different test cases."""
        for i, test_case in enumerate(TEST_CASES):
            self._check_test_case_fields(test_case, i)
            fmt = GerritReviewMessageFormatter(
                header=test_case["header"], footer=test_case["footer"]
            )
            for paragraph in test_case["paragraphs"]:
                fmt.append(paragraph)
            msg = fmt.format()
            self.assertEqual(
                msg,
                test_case["result"],
                "Formatted message does not match expected "
                "result in test case #%d:\n[%s]" % (i, msg),
            )


class TestGerritReview(unittest.TestCase):
    """Test that the GerritReview class behaves properly."""

    def test_str(self):
        """Test for str function."""
        obj = GerritReview()
        self.assertEqual(str(obj), "{}")

        obj2 = GerritReview(labels={"Verified": 1, "Code-Review": -1})
        self.assertEqual(str(obj2), '{"labels": {"Code-Review": -1, "Verified": 1}}')

        obj3 = GerritReview(
            comments=[{"filename": "Makefile", "line": 10, "message": "test"}]
        )
        self.assertEqual(
            str(obj3), '{"comments": {"Makefile": [{"line": 10, "message": "test"}]}}'
        )

        obj4 = GerritReview(
            labels={"Verified": 1, "Code-Review": -1},
            comments=[{"filename": "Makefile", "line": 10, "message": "test"}],
        )
        self.assertEqual(
            str(obj4),
            '{"comments": {"Makefile": [{"line": 10, "message": "test"}]},'
            ' "labels": {"Code-Review": -1, "Verified": 1}}',
        )

        obj5 = GerritReview(
            comments=[
                {"filename": "Makefile", "line": 15, "message": "test"},
                {"filename": "Make", "line": 10, "message": "test1"},
            ]
        )
        self.assertEqual(
            str(obj5),
            '{"comments": {"Make": [{"line": 10, "message": "test1"}],'
            ' "Makefile": [{"line": 15, "message": "test"}]}}',
        )


class TestNetrcAuth(unittest.TestCase):
    """Test that netrc authentication works."""

    def test_basic_auth_from_netrc(self):
        """Test that the HTTP basic auth is taken from netrc."""
        with patch("pygerrit2.rest.auth._get_netrc_auth") as mock_netrc:
            mock_netrc.return_value = ("netrcuser", "netrcpass")
            auth = HTTPBasicAuthFromNetrc(url="http://review.example.com")
            assert auth.username == "netrcuser"
            assert auth.password == "netrcpass"

    def test_digest_auth_from_netrc(self):
        """Test that the HTTP digest auth is taken from netrc."""
        with patch("pygerrit2.rest.auth._get_netrc_auth") as mock_netrc:
            mock_netrc.return_value = ("netrcuser", "netrcpass")
            auth = HTTPDigestAuthFromNetrc(url="http://review.example.com")
            assert auth.username == "netrcuser"
            assert auth.password == "netrcpass"

    def test_basic_auth_from_netrc_fails(self):
        """Test that an exception is raised when credentials are not found."""
        with self.assertRaises(ValueError) as exc:
            HTTPBasicAuthFromNetrc(url="http://review.example.com")
        assert str(exc.exception) == "netrc missing or no credentials found in netrc"

    def test_digest_auth_from_netrc_fails(self):
        """Test that an exception is raised when credentials are not found."""
        with self.assertRaises(ValueError) as exc:
            HTTPDigestAuthFromNetrc(url="http://review.example.com")
        assert str(exc.exception) == "netrc missing or no credentials found in netrc"

    def test_default_to_basic_auth_from_netrc(self):
        """Test auth defaults to HTTP basic from netrc when not specified."""
        with patch("pygerrit2.rest.auth._get_netrc_auth") as mock_netrc:
            mock_netrc.return_value = ("netrcuser", "netrcpass")
            api = GerritRestAPI(url="http://review.example.com")
            assert isinstance(api.auth, HTTPBasicAuthFromNetrc)
            assert api.url.endswith("/a/")

    def test_default_to_no_auth_when_not_in_netrc(self):
        """Test auth defaults to none when not specified and not in netrc."""
        with patch("pygerrit2.rest.auth._get_netrc_auth") as mock_netrc:
            mock_netrc.return_value = None
            api = GerritRestAPI(url="http://review.example.com")
            assert api.auth is None
            assert not api.url.endswith("/a/")

    def test_invalid_auth_type(self):
        """Test that an exception is raised for invalid auth type."""
        with self.assertRaises(ValueError) as exc:
            GerritRestAPI(url="http://review.example.com", auth="foo")
        assert re.search(r"Invalid auth type", str(exc.exception))

    def test_explicit_anonymous_with_netrc(self):
        """Test explicit anonymous access when credentials are in netrc."""
        with patch("pygerrit2.rest.auth._get_netrc_auth") as mock_netrc:
            mock_netrc.return_value = ("netrcuser", "netrcpass")
            auth = Anonymous()
            api = GerritRestAPI(url="http://review.example.com", auth=auth)
            assert api.auth is None
            assert not api.url.endswith("/a/")

    def test_explicit_anonymous_without_netrc(self):
        """Test explicit anonymous access when credentials are not in netrc."""
        with patch("pygerrit2.rest.auth._get_netrc_auth") as mock_netrc:
            mock_netrc.return_value = None
            auth = Anonymous()
            api = GerritRestAPI(url="http://review.example.com", auth=auth)
            assert api.auth is None
            assert not api.url.endswith("/a/")


class TestKwargsTranslation(unittest.TestCase):
    """Test that kwargs translation works."""

    def test_data_and_json(self):
        """Test that `json` and `data` cannot be used at the same time."""
        api = GerritRestAPI(url="http://review.example.com")
        with self.assertRaises(ValueError) as exc:
            api.translate_kwargs(data="d", json="j")
        assert re.search(r"Cannot use data and json together", str(exc.exception))

    def test_data_as_dict_converts_to_json_and_header_added(self):
        """Test that `data` dict is converted to `json`.

        Also test that a Content-Type header is added.
        """
        api = GerritRestAPI(url="http://review.example.com")
        data = {"a": "a"}
        result = api.translate_kwargs(data=data)
        assert "json" in result
        assert "data" not in result
        assert "headers" in result
        headers = result["headers"]
        assert "Content-Type" in headers
        assert result["json"] == {"a": "a"}
        assert headers["Content-Type"] == "application/json;charset=UTF-8"

    def test_json_is_unchanged_and_header_added(self):
        """Test that `json` is unchanged and a Content-Type header is added."""
        api = GerritRestAPI(url="http://review.example.com")
        json = {"a": "a"}
        result = api.translate_kwargs(json=json)
        assert "json" in result
        assert "data" not in result
        assert "headers" in result
        headers = result["headers"]
        assert "Content-Type" in headers
        assert result["json"] == {"a": "a"}
        assert headers["Content-Type"] == "application/json;charset=UTF-8"

    def test_json_no_side_effect_on_subsequent_call(self):
        """Test that subsequent call is not polluted with results of previous.

        If the translate_kwargs method is called, resulting in the content-type
        header being added, the header should not also be added on a subsequent
        call that does not need it.
        """
        api = GerritRestAPI(url="http://review.example.com")
        json = {"a": "a"}
        result = api.translate_kwargs(json=json)
        assert "json" in result
        assert "data" not in result
        assert "headers" in result
        headers = result["headers"]
        assert "Content-Type" in headers
        assert result["json"] == {"a": "a"}
        assert headers["Content-Type"] == "application/json;charset=UTF-8"
        kwargs = {"a": "a", "b": "b"}
        result = api.translate_kwargs(**kwargs)
        assert "json" not in result
        assert "data" not in result
        assert "a" in result
        assert "b" in result
        assert "headers" in result
        headers = result["headers"]
        assert "Content-Type" not in headers

    def test_kwargs_unchanged_when_no_data_or_json(self):
        """Test that `json` or `data` are not added when not passed."""
        api = GerritRestAPI(url="http://review.example.com")
        kwargs = {"a": "a", "b": "b"}
        result = api.translate_kwargs(**kwargs)
        assert "json" not in result
        assert "data" not in result
        assert "a" in result
        assert "b" in result
        assert "headers" in result
        headers = result["headers"]
        assert "Content-Type" not in headers

    def test_data_as_string_is_unchanged(self):
        """Test that `data` is unchanged when passed as a string."""
        api = GerritRestAPI(url="http://review.example.com")
        kwargs = {"data": "Content with non base64 valid chars åäö"}
        result = api.translate_kwargs(**kwargs)
        assert "json" not in result
        assert "data" in result
        assert result["data"] == "Content with non base64 valid chars åäö"
        assert "headers" in result
        headers = result["headers"]
        assert "Content-Type" not in headers


if __name__ == "__main__":
    unittest.main()



# The MIT License
#
# Copyright 2012 Sony Mobile Communications. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module to interface with Gerrit."""

from .rest import GerritRestAPI, GerritReview
from requests.auth import HTTPBasicAuth, HTTPDigestAuth
from .rest.auth import HTTPBasicAuthFromNetrc, HTTPDigestAuthFromNetrc, Anonymous

__all__ = [
    "Anonymous",
    "GerritRestAPI",
    "GerritReview",
    "HTTPBasicAuth",
    "HTTPDigestAuth",
    "HTTPBasicAuthFromNetrc",
    "HTTPDigestAuthFromNetrc",
]


def from_json(json_data, key):
    """Extract values from JSON data.

    :arg dict json_data: The JSON data
    :arg str key: Key to get data for.

    :Returns: The value of `key` from `json_data`, or None if `json_data`
        does not contain `key`.

    """
    if key in json_data:
        return json_data[key]
    return None


def escape_string(string):
    """Escape a string for use in Gerrit commands.

    :arg str string: The string to escape.

    :returns: The string with necessary escapes and surrounding double quotes
        so that it can be passed to any of the Gerrit commands that require
        double-quoted strings.

    """
    result = string
    result = result.replace("\\", "\\\\")
    result = result.replace('"', '\\"')
    return '"' + result + '"'


class GerritReviewMessageFormatter(object):
    """Helper class to format review messages that are sent to Gerrit.

    :arg str header: (optional) If specified, will be prepended as the first
        paragraph of the output message.
    :arg str footer: (optional) If specified, will be appended as the last
        paragraph of the output message.

    """

    def __init__(self, header=None, footer=None):
        """See class docstring."""
        self.paragraphs = []
        if header:
            self.header = header.strip()
        else:
            self.header = ""
        if footer:
            self.footer = footer.strip()
        else:
            self.footer = ""

    def append(self, data):
        """Append the given `data` to the output.

        :arg data: If a list, it is formatted as a bullet list with each
            entry in the list being a separate bullet.  Otherwise if it is a
            string, the string is added as a paragraph.

        :raises: ValueError if `data` is not a list or a string.

        """
        if not data:
            return

        if isinstance(data, list):
            # First we need to clean up the data.
            #
            # Gerrit creates new bullet items when it gets newline characters
            # within a bullet list paragraph, so unless we remove the newlines
            # from the texts the resulting bullet list will contain multiple
            # bullets and look crappy.
            #
            # We add the '*' character on the beginning of each bullet text in
            # the next step, so we strip off any existing leading '*' that the
            # caller has added, and then strip off any leading or trailing
            # whitespace.
            _items = [x.replace("\n", " ").strip().lstrip("*").strip() for x in data]

            # Create the bullet list only with the items that still have any
            # text in them after cleaning up.
            _paragraph = "\n".join(["* %s" % x for x in _items if x])
            if _paragraph:
                self.paragraphs.append(_paragraph)
        elif isinstance(data, str):
            _paragraph = data.strip()
            if _paragraph:
                self.paragraphs.append(_paragraph)
        else:
            raise ValueError("Data must be a list or a string")

    def is_empty(self):
        """Check if the formatter is empty.

        :Returns: True if empty, i.e. no paragraphs have been added.

        """
        return not self.paragraphs

    def format(self):
        """Format the message parts to a string.

        :Returns: A string of all the message parts separated into paragraphs,
            with header and footer paragraphs if they were specified in the
            constructor.

        """
        message = ""
        if self.paragraphs:
            if self.header:
                message += self.header + "\n\n"
            message += "\n\n".join(self.paragraphs)
            if self.footer:
                message += "\n\n" + self.footer
        return message



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License
#
# Copyright 2013 Sony Mobile Communications. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Example of using the Gerrit client REST API."""

import argparse
import logging
import sys

from requests.exceptions import RequestException

try:
    from requests_kerberos import HTTPKerberosAuth, OPTIONAL

    _KERBEROS_SUPPORT = True
except ImportError:
    _KERBEROS_SUPPORT = False

from pygerrit2 import GerritRestAPI
from pygerrit2 import HTTPDigestAuthFromNetrc, HTTPBasicAuthFromNetrc
from pygerrit2 import HTTPBasicAuth, HTTPDigestAuth


def _main():
    descr = "Send request using Gerrit HTTP API"
    parser = argparse.ArgumentParser(
        description=descr, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-g", "--gerrit-url", dest="gerrit_url", required=True, help="gerrit server url"
    )
    parser.add_argument(
        "-b",
        "--basic-auth",
        dest="basic_auth",
        action="store_true",
        help="(deprecated) use basic auth instead of digest",
    )
    parser.add_argument(
        "-d",
        "--digest-auth",
        dest="digest_auth",
        action="store_true",
        help="use digest auth instead of basic",
    )
    if _KERBEROS_SUPPORT:
        parser.add_argument(
            "-k",
            "--kerberos-auth",
            dest="kerberos_auth",
            action="store_true",
            help="use kerberos auth",
        )
    parser.add_argument("-u", "--username", dest="username", help="username")
    parser.add_argument("-p", "--password", dest="password", help="password")
    parser.add_argument(
        "-n",
        "--netrc",
        dest="netrc",
        action="store_true",
        help="Use credentials from netrc",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="enable verbose (debug) logging",
    )

    options = parser.parse_args()

    level = logging.DEBUG if options.verbose else logging.INFO
    logging.basicConfig(format="%(asctime)s %(levelname)s %(message)s", level=level)

    if _KERBEROS_SUPPORT and options.kerberos_auth:
        if options.username or options.password or options.basic_auth or options.netrc:
            parser.error(
                "--kerberos-auth may not be used together with "
                "--username, --password, --basic-auth or --netrc"
            )
        auth = HTTPKerberosAuth(mutual_authentication=OPTIONAL)
    elif options.username and options.password:
        if options.netrc:
            logging.warning("--netrc option ignored")
        if options.digest_auth:
            auth = HTTPDigestAuth(options.username, options.password)
        else:
            auth = HTTPBasicAuth(options.username, options.password)
    elif options.netrc:
        if options.digest_auth:
            auth = HTTPDigestAuthFromNetrc(url=options.gerrit_url)
        else:
            auth = HTTPBasicAuthFromNetrc(url=options.gerrit_url)
    else:
        auth = None

    rest = GerritRestAPI(url=options.gerrit_url, auth=auth)

    try:
        query = ["status:open"]
        if auth:
            query += ["owner:self"]
        else:
            query += ["limit:10"]
        changes = rest.get("/changes/?q=%s" % "%20".join(query))
        logging.info("%d changes", len(changes))
        for change in changes:
            logging.info(change["change_id"])
    except RequestException as err:
        logging.error("Error: %s", str(err))


if __name__ == "__main__":
    sys.exit(_main())



# The MIT License
#
# Copyright 2013 Sony Mobile Communications. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Authentication handlers."""

from requests.auth import HTTPDigestAuth, HTTPBasicAuth
from requests.utils import get_netrc_auth


def _get_netrc_auth(url):
    return get_netrc_auth(url)


class HTTPDigestAuthFromNetrc(HTTPDigestAuth):
    """HTTP Digest Auth with netrc credentials."""

    def __init__(self, url):
        """See class docstring."""
        auth = _get_netrc_auth(url)
        if not auth:
            raise ValueError("netrc missing or no credentials found in netrc")
        username, password = auth
        super(HTTPDigestAuthFromNetrc, self).__init__(username, password)


class HTTPBasicAuthFromNetrc(HTTPBasicAuth):
    """HTTP Basic Auth with netrc credentials."""

    def __init__(self, url):
        """See class docstring."""
        auth = _get_netrc_auth(url)
        if not auth:
            raise ValueError("netrc missing or no credentials found in netrc")
        username, password = auth
        super(HTTPBasicAuthFromNetrc, self).__init__(username, password)


class Anonymous:
    """No authentication; i.e. anonymous access."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License
#
# Copyright 2018 David Pursehouse. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Live server tests."""

import base64
import pytest
import unittest
from pygerrit2 import GerritRestAPI, GerritReview, HTTPBasicAuth, Anonymous
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_container_is_ready


TEST_TOPIC = "test-topic"


class GerritContainer(DockerContainer):
    """Gerrit container."""

    def __init__(self, version):
        """Construct a GerritContainer with the given version."""
        image = "gerritcodereview/gerrit:" + version
        super(GerritContainer, self).__init__(image)
        self.with_exposed_ports(8080)


@wait_container_is_ready()
def _initialize(api):
    api.get("/changes/")


@pytest.fixture(
    scope="module", params=["2.16.22", "3.0.13", "3.1.8", "3.2.3", "3.3.0-rc3"]
)
def gerrit_api(request):
    """Create a Gerrit container for the given version and return an API."""
    with GerritContainer(request.param) as gerrit:
        port = gerrit.get_exposed_port(8080)
        url = "http://localhost:%s" % port
        api = GerritRestAPI(url=url, auth=Anonymous())
        _initialize(api)
        auth = HTTPBasicAuth("admin", "secret")
        api = GerritRestAPI(url=url, auth=auth)
        yield api


class TestGerritAgainstLiveServer(object):
    """Run tests against a live server."""

    def _get_test_change(self, gerrit_api, topic=TEST_TOPIC):
        results = gerrit_api.get("/changes/?q=topic:" + topic)
        assert len(results) == 1
        return results[0]

    def test_put_with_json_dict(self, gerrit_api):
        """Test a PUT request passing data as a dict to `json`.

        Tests that the PUT request works as expected when data is passed
        via the `json` argument as a `dict`.

        Creates the test project which is used by subsequent tests.
        """
        projectinput = {"create_empty_commit": "true"}
        gerrit_api.put("/projects/test-project", json=projectinput)
        gerrit_api.get("/projects/test-project")

    def test_put_with_data_dict(self, gerrit_api):
        """Test a PUT request passing data as a dict to `data`.

        Tests that the PUT request works as expected when data is passed
        via the `data` argument as a `dict`.
        """
        description = {"description": "New Description"}
        gerrit_api.put("/projects/test-project/description", data=description)
        project = gerrit_api.get("/projects/test-project")
        assert project["description"] == "New Description"

    def test_post_with_data_dict_and_no_data(self, gerrit_api):
        """Test a POST request passing data as a dict to `data`.

        Tests that the POST request works as expected when data is passed
        via the `data` argument as a `dict`.
        """
        changeinput = {
            "project": "test-project",
            "subject": "subject",
            "branch": "master",
            "topic": "post-with-data",
        }
        result = gerrit_api.post("/changes/", data=changeinput)
        change = self._get_test_change(gerrit_api, "post-with-data")
        assert change["id"] == result["id"]

        # Subsequent post without data or json should not have the Content-Type
        # json header, and should succeed.
        result = gerrit_api.post("/changes/" + change["id"] + "/abandon")
        assert result["status"] == "ABANDONED"

    def test_post_with_json_dict(self, gerrit_api):
        """Test a POST request passing data as a dict to `json`.

        Tests that the POST request works as expected when data is passed
        via the `json` argument as a `dict`.

        Creates the change which is used by subsequent tests.
        """
        changeinput = {
            "project": "test-project",
            "subject": "subject",
            "branch": "master",
            "topic": TEST_TOPIC,
        }
        result = gerrit_api.post("/changes/", json=changeinput)
        change = self._get_test_change(gerrit_api)
        assert change["id"] == result["id"]

    def test_put_with_data_as_string(self, gerrit_api):
        """Test a PUT request passing data as a string to `data`.

        Tests that the PUT request works as expected when data is passed
        via the `data` parameter as a string.

        Creates a change edit that is checked in the subsequent test.
        """
        change_id = self._get_test_change(gerrit_api)["id"]
        gerrit_api.put(
            "/changes/" + change_id + "/edit/foo",
            data="Content with non base64 valid chars åäö",
        )

    def test_put_json_content(self, gerrit_api):
        """Test a PUT request with a json file content (issue #54)."""
        change_id = self._get_test_change(gerrit_api)["id"]
        content = """{"foo" : "bar"}"""
        gerrit_api.put("/changes/" + change_id + "/edit/file.json", data=content)

    def test_get_base64_data(self, gerrit_api):
        """Test a GET request on an API that returns base64 encoded response.

        Tests that the headers can be overridden on the GET call, resulting
        in the data being returned as text/plain, and that the content of the
        response can be base64 decoded.
        """
        change_id = self._get_test_change(gerrit_api)["id"]
        response = gerrit_api.get(
            "/changes/" + change_id + "/edit/foo", headers={"Accept": "text/plain"}
        )

        # Will raise binascii.Error if content is not properly encoded
        base64.b64decode(response)

    def test_get_patch_zip(self, gerrit_api):
        """Test a GET request to get a patch file (issue #19)."""
        change_id = self._get_test_change(gerrit_api)["id"]
        gerrit_api.get("/changes/" + change_id + "/revisions/current/patch?zip")

    def test_put_with_no_content(self, gerrit_api):
        """Test a PUT request with no content."""
        change_id = self._get_test_change(gerrit_api)["id"]
        gerrit_api.put("/changes/" + change_id + "/edit/foo")

    def test_review(self, gerrit_api):
        """Test that a review can be posted by the review API."""
        change_id = self._get_test_change(gerrit_api)["id"]
        review = GerritReview()
        review.set_message("Review from live test")
        review.add_labels({"Code-Review": 1})
        review.set_tag("a_test_tag")
        result = gerrit_api.review(change_id, "current", review)
        assert "labels" in result
        assert "Code-Review" in result["labels"]
        assert result["labels"]["Code-Review"] == 1


if __name__ == "__main__":
    unittest.main()



# The MIT License
#
# Copyright 2013 Sony Mobile Communications. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Interface to the Gerrit REST API."""

import json
import logging
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from .auth import HTTPBasicAuthFromNetrc, Anonymous

logger = logging.getLogger("pygerrit2")
fmt = "%(asctime)s-[%(name)s-%(levelname)s] %(message)s"
datefmt = "[%y-%m-%d %H:%M:%S]"
sh = logging.StreamHandler()
sh.setLevel(logging.WARNING)
sh.setFormatter(logging.Formatter(fmt, datefmt))
if not logger.handlers:
    logger.addHandler(sh)

GERRIT_MAGIC_JSON_PREFIX = ")]}'\n"
GERRIT_AUTH_SUFFIX = "/a"
DEFAULT_HEADERS = {"Accept": "application/json", "Accept-Encoding": "gzip"}


def _decode_response(response):
    """Strip off Gerrit's magic prefix and decode a response.

    :returns:
        Decoded JSON content as a dict, or raw text if content could not be
        decoded as JSON.

    :raises:
        requests.HTTPError if the response contains an HTTP error status code.

    """
    content_type = response.headers.get("content-type", "")
    logger.debug(
        "status[%s] content_type[%s] encoding[%s]"
        % (response.status_code, content_type, response.encoding)
    )
    response.raise_for_status()
    content = response.content.strip()
    if response.encoding:
        content = content.decode(response.encoding)
    if not content:
        logger.debug("no content in response")
        return content
    if content_type.split(";")[0] != "application/json":
        return content
    if content.startswith(GERRIT_MAGIC_JSON_PREFIX):
        index = len(GERRIT_MAGIC_JSON_PREFIX)
        content = content[index:]
    try:
        return json.loads(content)
    except ValueError:
        logger.error("Invalid json content: %s", content)
        raise


class GerritRestAPI(object):
    """Interface to the Gerrit REST API.

    :arg str url: The full URL to the server, including the `http(s)://`
        prefix. If `auth` is given, `url` will be automatically adjusted to
        include Gerrit's authentication suffix.
    :arg auth: (optional) Authentication handler.  Must be derived from
        `requests.auth.AuthBase`.
    :arg boolean verify: (optional) Set to False to disable verification of
        SSL certificates.
    :arg requests.adapters.BaseAdapter adapter: (optional) Custom connection
        adapter. See
        https://requests.readthedocs.io/en/master/api/#requests.adapters.BaseAdapter

    """

    def __init__(self, url, auth=None, verify=True, adapter=None):
        """See class docstring."""
        self.url = url.rstrip("/")
        self.session = requests.session()
        if not adapter:
            retry = Retry(
                total=5,
                read=5,
                connect=5,
                backoff_factor=0.3,
                status_forcelist=(500, 502, 504),
            )
            adapter = HTTPAdapter(max_retries=retry)
        self.session.mount("http://", adapter)
        self.session.mount("https://", adapter)

        if not auth:
            try:
                auth = HTTPBasicAuthFromNetrc(url)
            except ValueError as e:
                logger.debug("Error parsing netrc: %s", str(e))
                pass
        elif isinstance(auth, Anonymous):
            logger.debug("Anonymous")
            auth = None

        if auth:
            if not isinstance(auth, requests.auth.AuthBase):
                raise ValueError(
                    "Invalid auth type; must be derived from requests.auth.AuthBase"
                )

            if not self.url.endswith(GERRIT_AUTH_SUFFIX):
                self.url += GERRIT_AUTH_SUFFIX
        else:
            if self.url.endswith(GERRIT_AUTH_SUFFIX):
                self.url = self.url[: -len(GERRIT_AUTH_SUFFIX)]

        self.kwargs = {"auth": auth, "verify": verify}

        # Keep a copy of the auth, only needed for tests
        self.auth = auth

        if not self.url.endswith("/"):
            self.url += "/"

    def make_url(self, endpoint):
        """Make the full url for the endpoint.

        :arg str endpoint: The endpoint.

        :returns:
            The full url.

        """
        endpoint = endpoint.lstrip("/")
        return self.url + endpoint

    def translate_kwargs(self, **kwargs):
        """Translate kwargs replacing `data` with `json` if necessary."""
        local_kwargs = self.kwargs.copy()
        local_kwargs.update(kwargs)

        if "data" in local_kwargs and "json" in local_kwargs:
            raise ValueError("Cannot use data and json together")

        if "data" in local_kwargs and isinstance(local_kwargs["data"], dict):
            local_kwargs.update({"json": local_kwargs["data"]})
            del local_kwargs["data"]

        if "timeout" not in local_kwargs:
            local_kwargs.update({"timeout": 10})

        headers = DEFAULT_HEADERS.copy()
        if "headers" in kwargs:
            headers.update(kwargs["headers"])
        if "json" in local_kwargs:
            headers.update({"Content-Type": "application/json;charset=UTF-8"})
        local_kwargs.update({"headers": headers})

        return local_kwargs

    def get(self, endpoint, return_response=False, **kwargs):
        """Send HTTP GET to the endpoint.

        :arg str endpoint: The endpoint to send to.
        :arg bool return_response: If true will also return the response

        :returns:
            JSON decoded result.

        :raises:
            requests.RequestException on timeout or connection error.

        """
        args = self.translate_kwargs(**kwargs)
        response = self.session.get(self.make_url(endpoint), **args)

        decoded_response = _decode_response(response)

        if return_response:
            return decoded_response, response
        return decoded_response

    def put(self, endpoint, return_response=False, **kwargs):
        """Send HTTP PUT to the endpoint.

        :arg str endpoint: The endpoint to send to.

        :returns:
            JSON decoded result.

        :raises:
            requests.RequestException on timeout or connection error.

        """
        args = self.translate_kwargs(**kwargs)
        response = self.session.put(self.make_url(endpoint), **args)

        decoded_response = _decode_response(response)

        if return_response:
            return decoded_response, response
        return decoded_response

    def post(self, endpoint, return_response=False, **kwargs):
        """Send HTTP POST to the endpoint.

        :arg str endpoint: The endpoint to send to.

        :returns:
            JSON decoded result.

        :raises:
            requests.RequestException on timeout or connection error.

        """
        args = self.translate_kwargs(**kwargs)
        response = self.session.post(self.make_url(endpoint), **args)

        decoded_response = _decode_response(response)

        if return_response:
            return decoded_response, response
        return decoded_response

    def delete(self, endpoint, return_response=False, **kwargs):
        """Send HTTP DELETE to the endpoint.

        :arg str endpoint: The endpoint to send to.

        :returns:
            JSON decoded result.

        :raises:
            requests.RequestException on timeout or connection error.

        """
        args = self.translate_kwargs(**kwargs)
        response = self.session.delete(self.make_url(endpoint), **args)

        decoded_response = _decode_response(response)

        if return_response:
            return decoded_response, response
        return decoded_response

    def review(self, change_id, revision, review):
        """Submit a review.

        :arg str change_id: The change ID.
        :arg str revision: The revision.
        :arg str review: The review details as a :class:`GerritReview`.

        :returns:
            JSON decoded result.

        :raises:
            requests.RequestException on timeout or connection error.

        """
        endpoint = "changes/%s/revisions/%s/review" % (change_id, revision)
        return self.post(
            endpoint, data=str(review), headers={"Content-Type": "application/json"}
        )


class GerritReview(object):
    """Encapsulation of a Gerrit review.

    :arg str message: (optional) Cover message.
    :arg dict labels: (optional) Review labels.
    :arg dict comments: (optional) Inline comments.
    :arg str tag: (optional) Review tag.

    """

    def __init__(self, message=None, labels=None, comments=None, tag=None):
        """See class docstring."""
        self.message = message if message else ""
        self.tag = tag if tag else ""
        if labels:
            if not isinstance(labels, dict):
                raise ValueError("labels must be a dict.")
            self.labels = labels
        else:
            self.labels = {}
        if comments:
            if not isinstance(comments, list):
                raise ValueError("comments must be a list.")
            self.comments = {}
            self.add_comments(comments)
        else:
            self.comments = {}

    def set_message(self, message):
        """Set review cover message.

        :arg str message: Cover message.

        """
        self.message = message

    def set_tag(self, tag):
        """Set review tag.

        :arg str tag: Review tag.

        """
        self.tag = tag

    def add_labels(self, labels):
        """Add labels.

        :arg dict labels: Labels to add, for example

        Usage::

            add_labels({'Verified': 1,
                        'Code-Review': -1})

        """
        self.labels.update(labels)

    def add_comments(self, comments):
        """Add inline comments.

        :arg dict comments: Comments to add.

        Usage::

            add_comments([{'filename': 'Makefile',
                           'line': 10,
                           'message': 'inline message'}])

            add_comments([{'filename': 'Makefile',
                           'range': {'start_line': 0,
                                     'start_character': 1,
                                     'end_line': 0,
                                     'end_character': 5},
                           'message': 'inline message'}])

        """
        for comment in comments:
            if "filename" and "message" in list(comment.keys()):
                msg = {}
                if "range" in list(comment.keys()):
                    msg = {"range": comment["range"], "message": comment["message"]}
                elif "line" in list(comment.keys()):
                    msg = {"line": comment["line"], "message": comment["message"]}
                else:
                    continue
                file_comment = {comment["filename"]: [msg]}
                if self.comments:
                    if comment["filename"] in list(self.comments.keys()):
                        self.comments[comment["filename"]].append(msg)
                    else:
                        self.comments.update(file_comment)
                else:
                    self.comments.update(file_comment)

    def __str__(self):
        """Return a string representation."""
        review_input = {}
        if self.message:
            review_input.update({"message": self.message})
        if self.tag:
            review_input.update({"tag": self.tag})
        if self.labels:
            review_input.update({"labels": self.labels})
        if self.comments:
            review_input.update({"comments": self.comments})
        return json.dumps(review_input, sort_keys=True)
